from jinja2 import environment
from jinja2 import parser
from jinja2 import nodes
	
class Parser (parser.Parser):			
	def subparse(self, end_tokens=None):
		body = []
		data_buffer = []
		add_data = data_buffer.append

		if end_tokens is not None:
			self._end_token_stack.append(end_tokens)

		def flush_data():
			if data_buffer:
				lineno = data_buffer[0].lineno
				body.append(nodes.Output(data_buffer[:], lineno=lineno))
				del data_buffer[:]

		try:
			while self.stream:
				token = self.stream.current
				if token.type == 'data':
					if token.value:
						add_data(nodes.TemplateData(token.value,
													lineno=token.lineno))
					next(self.stream)
				elif token.type == 'variable_begin':
					next(self.stream)
					# added by Hans Roh 2016.5.1
					# If variable_begin & variable_end is #,
					# should excape by #"#"#
					# this patch make it easy to ##
					if self.stream.current.type == "variable_end":
						add_data(nodes.TemplateData(token.value, lineno=token.lineno))
						next(self.stream)
					else:	
						add_data(self.parse_tuple(with_condexpr=True))
						self.stream.expect('variable_end')
				elif token.type == 'block_begin':
					flush_data()
					next(self.stream)
					if end_tokens is not None and \
					   self.stream.current.test_any(*end_tokens):
						return body
					rv = self.parse_statement()
					if isinstance(rv, list):
						body.extend(rv)
					else:
						body.append(rv)
					self.stream.expect('block_end')
				else:
					raise AssertionError('internal parsing error')

			flush_data()
		finally:
			if end_tokens is not None:
				self._end_token_stack.pop()

		return body


class Environment (environment.Environment):
	def _parse(self, source, name, filename):
		return Parser(self, source, name, environment.encode_filename(filename)).parse()
	
def overlay (
	app_name, 
	variable_start_string = "{{", 
	variable_end_string = "}}",
	block_start_string = "{%", 
	block_end_string = "%}", 
	comment_start_string = "{#",
	comment_end_string = "#}",
	line_statement_prefix = "%", 
	line_comment_prefix = "%%",
	**karg
	):
	from jinja2 import PackageLoader
	
	if len (variable_start_string) == 1 and len (variable_end_string) == 1:
		env_class = Environment
	else:	
		env_class = environment.Environment
		
	return env_class (
		loader = PackageLoader (app_name),
		variable_start_string=variable_start_string,
		variable_end_string=variable_end_string,	  
		block_end_string = block_end_string,
		block_start_string = block_start_string,
		comment_start_string = comment_start_string,
		comment_end_string = comment_end_string,	  
		line_statement_prefix = line_statement_prefix,
		line_comment_prefix = line_comment_prefix,
		trim_blocks = True,
		lstrip_blocks = True,
		**karg
	)



import re, sys
from functools import wraps
from urllib.parse import unquote_plus, quote_plus, urljoin
import os
from aquests.lib import importer, strutil
from types import FunctionType as function
import inspect
from . import part
			
RX_RULE = re.compile ("(/<(.+?)>)")

class Saddlery (part.Part):
	def __init__ (self, *args, **kargs):
		part.Part.__init__ (self, *args, **kargs)
		self.packages = {}
		
	def mount (self, mount, module, partname = "app"):
		part = getattr (module, partname)
		part.init (module, partname, self.mount_p [:-1] + mount)
		self.packages [id (part)] = part
	
	def get_package_method (self, path_info, command, content_type, authorization, use_reloader = False):		
		app, method, kargs, options, match, matchtype = part.Part.get_package_method (
			self, path_info, command, content_type, authorization, use_reloader
		)
		
		if method is None and self.packages:
			for pid in list (self.packages.keys ()):
				package = self.packages [pid]
				subapp = getattr (package.module, package.packagename)
								
				if use_reloader and subapp.is_reloadable ():
					del self.packages [pid]
					args, its_packages = (package.mount_p, package.module, package.packagename), package.packages
					subapp.reload_package ()
					self.mount (*args)
					package.start (self.wasc, self.route, its_packages)
					subapp = getattr (package.module, package.packagename)
										
				app, method, kargs, options, match, matchtype = subapp.get_package_method (path_info, method, content_type, authorization, use_reloader)				
				if method:
					break
		
		if method is None:
			return None, None, None, None, None, 0
		
		if matchtype == -1: # 301 move
			return app, method, None, None, None, -1
			
		return app, [self._binds_request [0], method] + self._binds_request [1:4], kargs, options, match, matchtype
	
	def build_url (self, thing, *args, **kargs):
		url = part.Part.build_url (self, thing, *args, **kargs)
		if url is None:
			for p in self.packages:
				url = self.packages [p].build_url (thing, *args, **kargs)
				if url:
					return url
	
	def cleanup (self):
		# initing app & packages		
		self._binds_server [2] and self._binds_server [2] (self.wasc)
		for p in list (self.packages.values ()):
			p.cleanup ()
			
	def start (self, wasc, route, packages = None):
		self.wasc = wasc
		if not route: 
			self.route = "/"
		elif not route.endswith ("/"):			
			self.route = route + "/"
		else:
			self.route = route
			
		if packages is None:
			# initing app & packages
			self._binds_server [0] and self._binds_server [0] (self.wasc)
			for p in list (self.packages.values ()):
				p.start (self.wasc, route)
				
		else:
			self._binds_server [1] and self._binds_server [1] (self.wasc)
			self.packages = packages

	def restart (self, wasc, route):
		self.start (wasc, route, self.packages)
		


import pickle as pickle
from aquests.protocols.http import http_date
import time
import random
from aquests.lib import pathtool
from aquests.lib import udict
import os, sys
try:
	from urllib.parse import quote_plus, unquote_plus
except ImportError:
	from urllib import quote_plus, unquote_plus	
import base64
import pickle
from hmac import new as hmac
from . import named_session

NS_MBOX = "MBOX"
NS_SESSION = "SESS"

def crack_cookie (r):
	if not r: return {}
	arg={}
	q = [x.split('=', 1) for x in r.split('; ')]	
	for k in q:
		key = unquote_plus (k [0])
		if len (k) == 2:			
			if not key.startswith (NS_SESSION) and not key.startswith (NS_MBOX):
				arg[key] = unquote_plus (k[1])
			else:
				arg[key] = k[1]
		else:
			arg[key] = ""
	return arg


class BasicMethods:
	def __setitem__ (self, k, v):
		self.set (k, v)
		
	def __getitem__ (self, k):
		return self.get (k)
	
	def __delitem__ (self, k):
		return self.remove (k)
	
	def __contains__ (self, k):
		self.data is None and self.__parse ()
		return k in self.data
	
	def iterkeys (self):		
		self.data is None and self.__parse ()
		return self.data.iterkeys ()
	
	def itervalues (self):		
		self.data is None and self.__parse ()
		return self.data.itervalues ()	
	
	def iteritems (self):		
		self.data is None and self.__parse ()
		return self.data.iteritems ()	
		
	def has_key (self, k):
		self.data is None and self.__parse ()
		return k in self.data
	
	def items (self):
		self.data is None and self.__parse ()
		return list(self.data.items ())
	
	def keys (self):
		self.data is None and self.__parse ()
		return list(self.data.keys ())	
	
	def values (self):
		self.data is None and self.__parse ()
		return list(self.data.values ())	
		

class Cookie (BasicMethods):
	ACENTURY = 3153600000
	
	def __init__ (self, request, securekey = None, default_path = None, session_timeout = 1200):
		self.request = request		
		if securekey:
			self.securekey = securekey.encode ("utf8")
		else:
			self.securekey = securekey				
		self.default_path = default_path
		self.session_timeout = session_timeout
		self.dirty = False		
		self.data = None
		self.uncommits = {}
		self.sessions = {}		
	
	def __parse (self):
		self.data = {}	
		cookie = crack_cookie (self.request.get_header ("cookie"))
		for k, v in list(cookie.items ()):
			if k.startswith (NS_SESSION) or k.startswith (NS_MBOX):
				self.sessions [k] = v
				continue
			self.data [k] = v
		
	def get (self, k, a = None):
		self.data is None and self.__parse ()
		return self.data.get (k, a)
	
	def remove (self, k, path = None, domain = None):
		self.data is None and self.__parse ()
		try:
			del self.data [k]
		except KeyError:
			pass
		else:		
			self.set (k, "", 0, path, domain)
	
	def clear (self, path = None, domain = None):
		self.data is None and self.__parse ()
		for k, v in list(self.data.items ()):			
			self.set (k, "", 0, path, domain)
		self.data = {}
	
	def rollback (self):
		self.dirty = False		
		
	def commit (self):
		if self.data is None or not self.dirty:
			return			
		for cs in list(self.uncommits.values ()):
			self.request.response ["Set-Cookie"] = cs		
		self.dirty = False
		
	def set (self, name, val = "", expires = None, path = None, domain = None, secure = False, http_only = False):
		self.data is None and self.__parse ()
						
		self.dirty = True
		if path is None:
			path = self.default_path
		
		# browser string cookie
		cl = []
		if expires is not None:
			if expires == "never":
				expires = self.ACENTURY
			elif expires == "now":
				expires = 0

		if expires == 0:
			cl = ["%s=%s" % (name, "")]
			cl.append ("path=%s" % path)
		else:
			if name.startswith (NS_SESSION) or name.startswith (NS_MBOX):
				cl.append ("%s=%s" % (name, val))
			else:
				cl.append ("%s=%s" % (quote_plus (name), quote_plus (val)))
			cl.append ("path=%s" % path)		
		
		if expires is not None:
			cl.append ("expires=%s" % http_date.build_http_date (time.time () + expires))	
			
		if domain:
			cl.append ("domain=%s" % domain)			
		
		if secure:
			cl.append ("Secure")			
		
		if http_only:
			cl.append ("HttpOnly")
			
		self.uncommits [name] = "; ".join (cl)
		
		# cookie data
		if expires == 0:
			try: del self.data [name]
			except KeyError: pass										
		elif not name.startswith (NS_SESSION) and not name.startswith (NS_MBOX):
			self.data [name] = val
	
	def get_named_session_data (self, name):
		self.data is None and self.__parse ()
		return self.sessions.get (name)
				
	def get_session (self):
		return named_session.NamedSession ("session", self, self.request, self.securekey, self.session_timeout)
		
	def get_notices (self):
		return  named_session.NamedSession ("mbox", self, self.request, self.securekey, self.session_timeout)
		




#!/usr/bin/python3
# 2015. 11. 27 by Hans Roh hansroh@gmail.com

import sys, os, getopt
from skitai import lifetime
from aquests.lib import flock, pathtool, logger
from aquests.lib.athreads import select_trigger
from aquests.protocols.smtp import async_smtp, composer
import signal
import time
import glob
from skitai.server.wastuff import daemon

def hTERM (signum, frame):			
	lifetime.shutdown (0, 30.0)

def hKILL (signum, frame):			
	lifetime.shutdown (0, 1.0)

def hHUP (signum, frame):			
	lifetime.shutdown (3, 30.0)

class	SMTPDeliverAgent (daemon.Daemon):
	CONCURRENTS = 2
	MAX_RETRY = 10
	UNDELIVERS_KEEP_MAX = 2592000
	NAME = "smtpda"

	def __init__ (self, config, logpath, varpath, consol):
		self.que = {}
		self.actives = {}
		daemon.Daemon.__init__ (self, config, logpath, varpath, consol)
	
	def clean_shutdown_control (self, phase, time_in_this_phase):
		if phase == 1:
			self.shutdown_in_progress = True
	
	def close (self):
		self.logger ("service %s stopped" % self.NAME)
						
	def run (self):
		self.logger ("service %s started" % self.NAME)
		try:
			try:
				lifetime.loop (3.0)
			except KeyboardInterrupt:
				pass	
			except:
				self.logger.trace ()
		finally:
			self.close ()		
	
	def setup (self):
		self.make_logger ()
		self.bind_signal (hTERM, hKILL, hHUP)
		
		val = self.config.get ("max_retry", 10)
		if val: self.MAX_RETRY = val
		val = self.config.get ("keep_days", 30)
		if val: self.UNDELIVERS_KEEP_MAX = val * 3600 * 24

		self.default_smtp = (
			self.config.get ("smtpserver"),
			self.config.get ("user"),
			self.config.get ("password"),
			self.config.get ("ssl") and True or False
		)
		
		# dummy file like object for keeping lifetime loop
		if os.name == "nt":
			select_trigger.trigger.address = ('127.9.9.9', 19998)
		select_trigger.trigger ()
		
		self.path_spool = os.path.join (self.varpath, "smtpda",  "mail", "spool")
		pathtool.mkdir (self.path_spool)
		composer.Composer.SAVE_PATH = self.path_spool
		
		self.path_undeliver = os.path.join (self.varpath, "smtpda", "mail", "undeliver")
		pathtool.mkdir (self.path_undeliver)
		
		lifetime.init ()		
		lifetime.maintern.sched (2.0, self.handle_spool)		
		if os.name == "nt":			
			lifetime.maintern.sched (2.0, self.maintern_shutdown_request)		
		
	def send (self):
		if self.shutdown_in_progress:
			return
			
		while self.que:
			fn = self.que.popitem () [0]
			if fn in self.actives: continue

			self.actives [fn] = time.time ()
			path = os.path.join (self.path_spool, fn)
			cmps = composer.load (path)
			if cmps.get_SMTP () is None:		
				cmps.set_smtp (*self.default_smtp)
			if cmps.is_SSL ():
				request = async_smtp.SMTP_SSL
			else:	
				request = async_smtp.SMTP
			request (cmps, self.logger, self.when_done)					
			break
	
	def when_done (self, cmps, code, reps):
		fn = os.path.split (cmps.get_FILENAME ()) [-1]
		
		if code == -250:
			cmps.remove ()
		else:
			if cmps.get_RETRYS () > self.MAX_RETRY:
				cmps.moveto (self.path_undeliver)
			else:	
				cmps.moveto (self.path_spool)
		
		del self.actives [fn]
		self.send ()
	
	def maintern (self, current_time):
		for path in glob.glob (os.path.join (self.path_undeliver, "*.*")):
			mtime = os.path.getmtime (path)
			if mtime + self.UNDELIVERS_KEEP_MAX > current_time: # over a month
				try: os.remove (path)
				except: self.logger.trace ()
		self.last_maintern = time.time ()
		
	def handle_spool (self, current_time):
		if self.shutdown_in_progress:
			return
		
		if self.last_maintern + 3600 >  current_time:
			self.maintern (current_time)
		
		if not self.que: # previous que has priority			
			for path in glob.glob (os.path.join (self.path_spool, "*.*")):
				try:
					fn = os.path.split (path)[-1]
					retrys, digest = fn.split (".")
					assert (len (digest) == 32)
					retrys = int (retrys)
				except (ValueError, AssertionError):
					continue
				except:
					self.logger.trace ()
					continue
						
				mtime = os.path.getmtime (path)
				if retrys > 7:
					delta = 21600
				elif retrys == 0:
					delta = 0
				else:
					delta = 4 ** retrys
				
				if retrys >= self.MAX_RETRY:
					self.que [fn] = None
				elif mtime + delta <= current_time and fn not in self.que and fn not in self.actives:
					self.que [fn] = None
		
		for i in range (self.CONCURRENTS - len (self.actives)):
			self.send ()
	
		
def usage ():
		print("""
Usage:
	smtpda.py [options...]

Options:	
	--var-path=
	--log-path=
	--verbose or -v
	--help or -h

Examples:	
	ex. smtpda.py -v
	""")


if __name__ == "__main__":
	argopt = getopt.getopt(
		sys.argv[1:], 
		"hv", 
		[
			"help", "verbose=", "log-path=", "var-path=", "pname=",
			"max-retry=", "keep-days=", "server=", "user=", "password=", "ssl="
		]
	)
	_consol = "no"
	_cf = {
		"max-retry": 3,
		"keep-days": 3,	
		"ssl": 0,
	}
	_logpath, _varpath = None, None
	for k, v in argopt [0]:
		if k == "--help" or k == "-h":
			usage ()
			sys.exit ()		
		elif k == "--verbose" or k == "-v":
			_consol = v
		elif k == "--log-path":	
			_logpath = v
		elif k == "--var-path":	
			_varpath = v
		elif k == "--max-retry":	
			_cf ['max-retry'] = int (v)
		elif k == "--keep-days":	
			_cf ['keep-days'] = int (v)	
		elif k == "--server":	
			_cf ['smtpserver'] = v
		elif k == "--user":	
			_cf ['user'] = v
		elif k == "--password":	
			_cf ['password'] = v	
		elif k == "--ssl":	
			_cf ['ssl'] = 1		
		elif k == "--pname":	
			_cf ['pname'] = v
				
	service = daemon.make_service (SMTPDeliverAgent, _cf, _logpath, _varpath, _consol)
	try:
		service.run ()		
	finally:	
		if _consol not in ("1", "yes"):
			sys.stderr.close ()	
		sys.exit (lifetime._exit_code)



from skitai.server import http_response
from aquests.lib import producers
import time


class response (http_response.http_response):	
	def __init__ (self, request):
		http_response.http_response.__init__ (self, request)
	
	def set_streaming (self):
		self._is_async_streaming = True
		
	def build_reply_header (self):	
		h = [(b":status", str (self.reply_code).encode ("utf8"))]
		for k, v in self.reply_headers:
			h.append ((k.encode ("utf8"), str (v).encode ("utf8")))
		return h
	
	def hint_promise (self, uri):
		headers = [
			(':path', uri),
			(':authority', self.request.get_header ('host')),	    
	    (':scheme', self.request.scheme),
	    (':method', "GET")	    
	  ]
	   
		additional_headers = [
	   	(k, v) for k, v in [
		   	('accept-encoding', self.request.get_header ('accept-encoding')),
		   	('accept-language', self.request.get_header ('accept-language')),
		    ('cookie', self.request.get_header ('cookie')),
		    ('user-agent', self.request.get_header ('user-agent')),
		    ('referer', "%s://%s%s" % (self.request.scheme, self.request.get_header ('host'), self.request.uri)),	    
		   ] if v
	   ]
	    
		self.request.http2.push_promise (self.request.stream_id, headers, additional_headers)
		
	def done (self, force_close = False, upgrade_to = None):
		self.content_type = self.get ('content-type')
		
		if not self.is_responsable (): return
		self._is_done = True
		if self.request.http2 is None: return
		
		self.htime = (time.time () - self.stime) * 1000
		self.stime = time.time () #for delivery time
		
		# removed by HTTP/2.0 Spec.
		self.delete ('transfer-encoding')
		self.delete ('connection')
		
		# compress payload and globbing production
		do_optimize = True
		if upgrade_to or self.is_async_streaming ():
			do_optimize = False
		
		if not self.outgoing:
			self.delete ('content-type')
			self.delete ('content-length')
			outgoing_producer = None
			
		elif len (self.outgoing) == 1 and hasattr (self.outgoing.first (), "ready"):
			outgoing_producer = producers.composite_producer (self.outgoing)
			do_optimize = False			
			
		else:
			outgoing_producer = producers.composite_producer (self.outgoing)			
			if do_optimize and not self.has_key ('Content-Encoding'):
				way_to_compress = ""			
				maybe_compress = self.request.get_header ("Accept-Encoding")
				
				if maybe_compress:
					cl = self.has_key ("content-length") and int (self.get ("Content-Length")) or -1
					if cl == -1:
						cl = self.outgoing.get_estimate_content_length ()					
					if 0 < cl <= http_response.UNCOMPRESS_MAX:
						maybe_compress = ""
					
				if maybe_compress:
					content_type = self.get ("Content-Type")
					if maybe_compress and content_type and (content_type.startswith ("text/") or content_type.startswith ("application/json")):
						accept_encoding = [x.strip () for x in maybe_compress.split (",")]
						if "gzip" in accept_encoding:
							way_to_compress = "gzip"
						elif "deflate" in accept_encoding:
							way_to_compress = "deflate"
			
				if way_to_compress:
					if self.has_key ('Content-Length'):
						self.delete ("content-length") # rebuild
					self.update ('Content-Encoding', way_to_compress)			
					if way_to_compress == "gzip":
						compressing_producer = producers.gzipped_producer
					else: # deflate
						compressing_producer = producers.compressed_producer		
					outgoing_producer = compressing_producer (outgoing_producer)
			
		if self.request.http2 is None: return

		if upgrade_to:
			# do not change http2 channel
			request, terminator = upgrade_to
			self.request.channel.current_request = request
			self.request.channel.set_terminator (terminator)
		
		logger = self.request.logger #IMP: for  disconnect with request
		try:
			self.request.http2.handle_response (
				self.request.stream_id, 
				self.build_reply_header (),
				outgoing_producer,
				do_optimize,				
				force_close = force_close
			)
			
		except:
			logger.trace ()			
		


#!/usr/bin/python3
# 2016. 2. 29 by Hans Roh hansroh@gmail.com

import subprocess
import sys, os, getopt
from aquests.lib import flock, pathtool, logger, confparse, killtree, processutil
import signal
import time
import glob
import threading
from datetime import datetime
from skitai.server.wastuff import daemon


def hTERM (signum, frame):			
	daemon.EXIT_CODE = 0

def hHUP (signum, frame):			
	daemon.EXIT_CODE = 3

class	CronManager (daemon.Daemon):	
	NAME = "cron"
	
	def __init__ (self, config, logpath, varpath, consol):
		self.jobs = []
		self.currents = {}		
		daemon.Daemon.__init__ (self, config, logpath, varpath, consol)
		
	def close (self):
		self.logger ("trying to kill childs...")
		self.maintern (time.time (), kill = True)
		self.logger ("service cron stopped")
	
	def kill (self, pid, cmd):
		self.logger ("trying to kill pid:%d %s" % (pid, cmd))		
		if processutil.is_running (pid)	:
			killtree.kill (pid, True)
			
	def maintern (self, current_time, kill = False):
		for pid, (cmd, started, child) in list (self.currents.items ()):			
			rcode = child.poll ()
			due = ((current_time - started) * 1000)
			
			if due > 60000:
				due -= 60000				
			if due > 120000:
				due = "%2.1f min" % (due / 1000. / 60.,)
			elif due > 1000:
				due = "%2.1f sec" % (due / 1000.,)
			else:
				due = "%d ms" % due				
			
			if rcode is None:
				if kill:
					 self.kill (pid, cmd)					 
					 del self.currents [pid]					 
				else:
					self.logger ("job still running pid:%d for %s, %s" % (pid, due, cmd))
				continue
			
			self.logger ("job has been finished pid:%d with rcode:%d for %s, %s" % (pid, rcode, due, cmd))
			del self.currents [pid]
		self.last_maintern = time.time ()

	def run (self):
		self.logger ("service cron started")
		try:
			try:
				self.loop ()
			except KeyboardInterrupt:
				pass
			except:
				self.logger.trace ()
		finally:
			self.close ()		
	
	def parse (self, unitd, umax = 60):
		#print ("---", unitd)
		
		if unitd == "*":
			return []
						
		try: 
			unit, interval = unitd.split ("/", 1)
		except ValueError:
			unit, interval = unitd, 0
		else:
			interval = int (interval)
			if interval == 1 and unit in "0*":
				return []			
			if unit == "*":
				unit = "0"								
			if interval > umax / 2.:
				raise ValueError ("interval %d is too big" % interval)
		
		units = {}
		_all = unit.split (",")
		_all_len = len (_all)
		
		for subunit in _all:
			if subunit.find ("-") == -1:
				if interval == 0 or _all_len > 1:
					units [int (subunit)] = None
				else:	
					for i in range (int (subunit), umax, interval):
						units [i] = None
						
			else:
				a, b = subunit.split ("-", 1)
				if a == "":
					a = 0
				else:
					a = int (a)
				if b == "":
					b = umax			
				else:	
					b = int (b) + 1
				for i in range (a, b, interval == 0 and 1 or interval):
					units [i] = None
		
		# add sunday (0 or 7)
		if umax == 7 and 0 in units:
			del units [0]
			units [7] = None
			
		r = list (units.keys ())
		r.sort ()
		return r
	
	def update_jobs (self, jobs):			
		self.jobs = []				
		
		if not jobs: return
		for job in jobs:
			args = job.split (" ", 5)
			if len (args) != 6:
				self.logger ("invalid cron command %s" % (args,), "error")
				continue
			
			try:
				sched = (
					self.parse (args [0], 60),
					self.parse (args [1], 24),
					self.parse (args [2], 31),
					self.parse (args [3], 12),
					self.parse (args [4], 7),
					args [5]
				)					
			
			except ValueError as why:
				self.logger ("%s, %s" % (job, why), "error")
				continue	
				
			except:
				self.logger.trace (job)
				continue	
			
			else:
				self.jobs.append (sched)
				self.logger ("job added %s" % job)
	
	def check_sched_per_minute (self):				
		now = datetime.now ().timetuple ()	
		for m, h, d, M, w, cmd in self.jobs:			
			if M and now.tm_mon not in M:
				continue
			if w and now.tm_wday + 1 not in w:
				continue
			if d and now.tm_mday not in d:
				continue	
			if h and now.tm_hour not in h:
				continue
			if m and now.tm_min not in m:
				continue
			else:
				threading.Thread (target = self.execute, args = (cmd,)).start ()
		
	def setup (self):
		self.make_logger ()
		self.bind_signal (hTERM, hTERM, hHUP)		
		self.update_jobs (self.config.get ('jobs', []))
		
	def execute (self, cmd):					
		if os.name == "nt":
			child = subprocess.Popen (
				cmd, 
				#shell = True, if true, can't terminate process bacuase run within cmd's child
				creationflags=subprocess.CREATE_NEW_PROCESS_GROUP
			)

		else:
			child = subprocess.Popen (
				"exec " + cmd, 
				shell = True
			)
		
		self.logger ("job started with pid:%d %s" % (child.pid, cmd))
		self.currents [child.pid] = (cmd, time.time (), child)
	
	def loop (self):
		while 1:
			if daemon.EXIT_CODE is not None:
				break				
			now = time.time ()			
			self.maintern (now)
			self.check_sched_per_minute ()
			for i in range (60):
				if os.name == "nt" and i % 10 == 0:
					self.maintern_shutdown_request (now)
				if daemon.EXIT_CODE is not None:
					break
				time.sleep (1)
		

def usage ():
		print("""
Usage:
	cron.py [options...] [jobs...]

Options:
	--var-path=
	--log-path=
	--verbose or -v
	--help or -h

Examples:
	ex. cron.py -v	
	""")


if __name__ == "__main__":
	argopt = getopt.getopt(
		sys.argv[1:], 
		"hv", 
		[
			"help", "verbose=", "log-path=", "var-path=", "pname="
		]	
	)
	
	_consol = "no"
	_cf = {}
	_logpath, _varpath = None, None
	
	for k, v in argopt [0]:
		if k == "--help" or k == "-h":	
			usage ()
			sys.exit ()		
		elif k == "--verbose" or k == "-v":
			_consol = v
		elif k == "--log-path":	
			_logpath = v
		elif k == "--var-path":	
			_varpath = v
		elif k == "--pname":	
			_cf ["pname"] = v
			
	_cf ['jobs'] = []
	for job in argopt [1]:
		_cf ['jobs'].append (job)
		
	service = daemon.make_service (CronManager, _cf, _logpath, _varpath, _consol)	
	try:
		service.run ()
	finally:	
		if _consol not in ("1", "yes"):
			sys.stderr.close ()
		sys.exit (daemon.EXIT_CODE)
		